import struct
from qmdl.anorms import normals


# This wrapper function keeps the file reading code tidy
def _struct_read(s, file):
    return s.unpack(file.read(s.size))

class ParseError(BaseException):
    pass
class Mdl:
    # STRUCTS
    # A bunch of static objects to avoid repeated construction during loops
    header_struct = struct.Struct("<4sl3f3ff3fllllllllf")
    single_struct = struct.Struct("<l")
    duration_struct = struct.Struct("<f")
    triple_struct = struct.Struct("<3l")
    coord_struct = struct.Struct("<4B")
    name_struct = struct.Struct("<16s")

    # INIT
    # Fill in defaults everywhere to avoid AtrributeError and create any lists
    def __init__(self):
        # initialise all the exported properties to minimal defaults
        self.ident = b"IDPO"
        self.version = 6
        self.scale = (0, 0, 0)
        self.origin = (0, 0, 0)
        self.boundingradius = 0
        self.eyeposition = (0, 0, 0)
        self.skins = []
        self.skinwidth, self.skinheight = 0, 0
        self.vertices = []
        self.triangles = []
        self.frames = []
        self.synctype = 1
        self.flags = 0
        self.size = 0

    # HEADER
    # This is stored against the body of the Mdl object
    # a) because you can only have a single header per model and
    # b) because quantities like num_skins only make sense in terms of an
    #   actual collection of skins belonging to the model
    # Note: private fields like _num_verts are only used to store the lengths
    # of field parsed from the header of a file we are current reading
    # Once these collections are read from the binary file, we should get the
    # size of collections from len(self.vertices) etc. as per the write_header
    # function below
    def read_header(self, file):
        (self.ident, self.version,
         scale_x, scale_y, scale_z,
         origin_x, origin_y, origin_z,
         self.boundingradius,
         eyeposition_x, eyeposition_y, eyeposition_z,
         self._num_skins,
         self.skinwidth, self.skinheight,
         self._num_verts,
         self._num_tris,
         self._num_frames,
         self.synctype,
         self.flags,
         self.size
         ) = _struct_read(Mdl.header_struct, file)
        #run some basic checks that we have been given a model file
        if self.ident != b"IDPO":
            raise ParseError("Ident does not match IDPO")
        if self.version != 6:
            raise ParseError("mdl file is not version 6")
        if self.skinwidth <= 0:
            raise ParseError("mdl file has invalid skinwidth")
        if self.skinheight <= 0:
            raise ParseError("mdl file has invalid skinheight")
        #package the vectors into tuples
        self.scale = (scale_x, scale_y, scale_z)
        self.origin = (origin_x, origin_y, origin_z)
        self.eyeposition = (eyeposition_x, eyeposition_y, eyeposition_z)
        return self

    def write_header(self, file):
        file.write(Mdl.header_struct.pack
                   (
                       self.ident, self.version,
                       self.scale[0], self.scale[1], self.scale[2],
                       self.origin[0], self.origin[1], self.origin[2],
                       self.boundingradius,
                       self.eyeposition[0], self.eyeposition[1],
                           self.eyeposition[2],
                       len(self.skins),
                       self.skinwidth, self.skinheight,
                       len(self.vertices),
                       len(self.triangles),
                       len(self.frames),
                       self.synctype,
                       self.flags,
                       self.size)
                   )
        return self

    # SKIN
    # Class representing a quake skin
    # Pixel data is stored as a flat binary string
    # Object also has height and width of skin stored
    # 'pixel' helper function will return the palette index at x,y coordinates
    class Skin:
        def __init__(self, width, height):
            self.width = width
            self.height = height
            self.pixels = b"\x00" * width * height

        def read(self, file):
            self.pixels = file.read(self.width * self.height)
            return self

        def write(self, file, write_header=True):
            # Skingroups will call this on each skin they contain when written,
            # but with write_header set to false, as they write a single
            # header for the entire skingroup beforehand
            if write_header:
                file.write(Mdl.single_struct.pack(0))
            file.write(self.pixels)
            return self

        def pixel(self, x, y):
            x = x % self.width
            y = y % self.height
            return self.pixels[x + y * self.height]


    # SKINGROUP
    # For grouped animated skins
    # Comprises a header including a list of durations, followed by a list of
    # Skin objects.
    # Unenforced constraint: the list of durations and skins must be equal in
    # length or the mdl file will be invalid
    class SkinGroup:
        def __init__(self, width, height):
            self.width = width
            self.height = height
            self.duration = []
            self.skins = []

        def read(self, file):
            count = _struct_read(Mdl.single_struct, file)[0]
            for i in range(count):
                self.duration.append(
                    _struct_read(Mdl.duration_struct, file)[0])
            for i in range(count):
                self.skins.append(Mdl.Skin(self.width, self.height).read(file))
            return self

        def write(self, file):
            file.write(Mdl.single_struct.pack(1))
            file.write(Mdl.single_struct.pack(len(self.skins)))
            for dur in self.duration:
                file.write(Mdl.duration_struct.pack(dur))
            for s in self.skins:
                s.write(file, write_header=False)
            return self
        
    # READ/WRITE SKINS
    # These functions are methods of the main Mdl class
    # They read to/write from the skins list on the Mdl object
    # Note that the list can contain a mixture of Skin and Skingroup objects
    # When writing we can treat each member identically by calling the object's
    # own 'write' function
    def read_skins(self, file):
        for i in range(self._num_skins):
            type = _struct_read(Mdl.single_struct, file)[0]
            if type == 0:
                skin = Mdl.Skin(self.skinwidth, self.skinheight)
            else:
                skin = Mdl.SkinGroup(self.skinwidth, self.skinheight)
            self.skins.append(skin.read(file))
        return self

    def write_skins(self, file):
        for skin in self.skins:
            skin.write(file)
        return self
            

    # VERTEX
    # This class stores the skin coordinates for each vertex in the model
    # The vertex positions in the poses of the model are stored in the frame
    # classes. The triangles refer to the vertices by index of their position
    # in this list.
    # Unenforced constraints: The vertex list must remain in sync with the
    # triangle and frame list. For example: removing a vertex from this list
    # requires you to remove the vertex from every pose in the model frames.
    # In addition you must update the triangles list, not just to remove every
    # triangle which contained the vertex. You must also to reduce by one the
    # index values recorded in each triangle of every vertex which follows the
    # deleted vertex, since each will have shifted one place down the list.
    class Vertex:
        def __init__(self):
            self.onseam, self.u, self.v = 0, 0, 0
            
        def read(self, file):
            self.onseam, self.u, self.v = _struct_read(Mdl.triple_struct, file)
            return self

        def write(self, file):
            file.write(Mdl.triple_struct.pack(self.onseam, self.u, self.v))
            return self
        
    # READ/WRITE VERTICES
    # These functions are methods of the main Mdl class
    # They operate on the 'vertices' list it maintains
    def read_vertices(self, file):
        for i in range(self._num_verts):
            self.vertices.append(Mdl.Vertex().read(file))
        return self

    def write_vertices(self, file):
        for vertex in self.vertices:
            vertex.write(file)
        return self

    # TRIANGLE
    # This class represents a triangle in the model as it is stored in the mdl
    # format, a backfacing flag followed by three indicies into the vertex list.
    # See VERTEX for the details of the relation between these classes and FRAME
    class Triangle:
        def __init__(self):
            self.backface, self.vertices = 0, (0, 0, 0)
            
        def read(self, file):
            self.backface = _struct_read(Mdl.single_struct, file)[0]
            self.vertices = _struct_read(Mdl.triple_struct, file)
            return self
            
        def write(self, file):
            file.write(Mdl.single_struct.pack(self.backface)+
                       Mdl.triple_struct.pack(*self.vertices)
                       )
            return self

    # READ/WRITE TRIANGLES
    # These functions are methods of the main Mdl class
    # They operate on the 'triangles' list it maintains
    def read_triangles(self, file):
        for i in range(self._num_tris):
            self.triangles.append(Mdl.Triangle().read(file))
        return self

    def write_triangles(self, file):
        for triangle in self.triangles:
            triangle.write(file)
        return self
            
    # COORD
    # This class stores the 'compressed' coordinate the mdl file uses for vertex
    # position. x, y, z values are stored as a byte in the range 0-255. Vertex
    # normals are stored in a 4th byte in the range 0-161. The normal value is
    # a lookup in the standard table of normals provided in anorms.h
    # This class contains a pair of helper functions to deal with the normals
    # 'decode' returns the normal vector that the coord stores
    # 'encode' takes a vector and sets the normal of the coord to the index of
    # the standard normal which best approximates its direction.
    # See also TRANSLATE below.
    class Coord:
        def __init__(self):
            self.position, self.normal = (0, 0, 0), 0
        
        def read(self, file):
            coord = _struct_read(Mdl.coord_struct, file)
            self.position = coord[0:3]
            self.normal = coord[3]
            return self
        
        def write(self, file):
            file.write(Mdl.coord_struct.pack(*(self.position + (self.normal,))))
            return self
        
        def encode(self, vector):
            m = 0
            n = 0
            for i in range(len(normals)):
                #dot = sum(a * b for a, b in zip(vector, normals[i]))
                dot = vector[0] * normals[i][0] + vector[1] * normals[i][1] +\
                      vector[2] * normals[i][2] 
                if dot > m:
                    m = dot
                    n = i
            self.normal = n
            return self

        def decode(self):
            return normals[self.normal]
        
    # TRANSLATE
    # This is another helper function for a Coord, but it is a member of the Mdl
    # object. It translates the 0-255 range coords into floating point 'world
    # coordinates'. This method belongs to the Mdl itself because we need to use
    # the origin and scale properties of the Mdl to interpret the Coord.
    def translate(self, coord):
        return tuple(org + scale * x for org, scale, x in zip(
            self.origin, self.scale, coord.position))
            
    class Frame:
        def __init__(self):
            self.bbox_min = Mdl.Coord()
            self.bbox_max = Mdl.Coord()
            self.name = b"nameless"
            self.vertices = []
            
        def read(self, file, vertcount):
            self.bbox_min.read(file)
            self.bbox_max.read(file)
            self.name = file.read(16).split(b"\x00")[0]
            for i in range(vertcount):
                self.vertices.append(Mdl.Coord().read(file))
            return self
                
        def write(self, file, write_header=True):
            if write_header:
                file.write(Mdl.single_struct.pack(0))
            self.bbox_min.write(file)
            self.bbox_max.write(file)
            file.write(Mdl.name_struct.pack(self.name))
            for vert in self.vertices:
                vert.write(file)
            return self
                
    class FrameGroup:
        def __init__(self):
            self.sequencemin = Mdl.Coord()
            self.sequencemax = Mdl.Coord()
            self.duration = []
            self.frames = []
                          
        def read(self, file, vertcount):
            count = _struct_read(Mdl.single_struct, file)[0]
            self.sequencemin.read(file)
            self.sequencemax.read(file)
            for i in range(count):
                self.duration.append(_struct_read(Mdl.duration_struct, file)[0])
            for i in range(count):
                self.frames.append(Mdl.Frame().read(file, vertcount))
            return self
        
        def calculate_bounds(self):
            _sequencemin = (255, 255, 255)
            _sequencemax = (0, 0, 0)
            for frame in self.frames:
                _bbox_min = frame.bbox_min.position
                _bbox_max = frame.bbox_max.position
                _sequencemin = (min(_sequencemin[0], _bbox_min[0]),
                                min(_sequencemin[1], _bbox_min[1]),
                                min(_sequencemin[2], _bbox_min[2]))
                _sequencemax = (max(_sequencemax[0], _bbox_max[0]),
                                max(_sequencemax[1], _bbox_max[1]),
                                max(_sequencemax[2], _bbox_max[2]))
            self.sequencemin.position = _sequencemin
            self.sequencemax.position = _sequencemax

        def write(self, file):
            file.write(Mdl.single_struct.pack(1))
            file.write(Mdl.single_struct.pack(len(self.frames)))
            self.calculate_bounds()
            self.sequencemin.write(file)
            self.sequencemax.write(file)
            for dur in self.duration:
                file.write(Mdl.duration_struct.pack(dur))
            for frame in self.frames:
                frame.write(file, write_header=False)
            return self

        


    def read_frames(self, file):
        for i in range(self._num_frames):
            type = _struct_read(Mdl.single_struct, file)[0]
            if type == 0:
                frame = Mdl.Frame()
            else:
                frame = Mdl.FrameGroup()
            self.frames.append(frame.read(file, self._num_verts))
        return self
            
    def write_frames(self, file):
        for frame in self.frames:
            frame.write(file)
        return self


    def read(self, file):
        self.read_header(file)
        self.read_skins(file)
        self.read_vertices(file)
        self.read_triangles(file)
        self.read_frames(file)
        return self
        
    def write(self, file):
        self.write_header(file)
        self.write_skins(file)
        self.write_vertices(file)
        self.write_triangles(file)
        self.write_frames(file)
        return self


